==> HDPenReg_0.89.5.tar.gz: HDPenReg/src/stkpp/projects/Analysis/include/STK_Algo_FindZero.h <==
/*--------------------------------------------------------------------*/
/*     Copyright (C) 2004-2007  Serge Iovleff

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place,
    Suite 330,
    Boston, MA 02111-1307
    USA

    Contact : S..._Dot_I..._At_stkpp_Dot_org (see copyright for ...)
*/

/*
 * Project:  stkpp::STatistiK
 * Purpose:  Method implementing Algorithms.
 * Author:   Serge Iovleff, S..._Dot_I..._At_stkpp_Dot_org (see copyright for ...)
 **/

/** @file STK_Algo.h
 *  @brief In this file we give some templated generic algorithms.
 **/

#ifndef STK_ALGO_FINDZERO_H
#define STK_ALGO_FINDZERO_H

#include "../../STKernel/include/STK_Integer.h"
#include "../../STKernel/include/STK_Real.h"
#include "../../STKernel/include/STK_Macros.h"
#include "STK_IFunction.h"

#define MAX_ITER 1000000

namespace STK
{

namespace Algo
{
// forward declaration
template <class Function>
Real BrentMethod( IFunction<Function> const& f, Real const& x0, Real const& x1, Real tol);

/** @ingroup Analysis
 *  @brief apply the secant method for finding the zero of a function.
 *
 * Compute iteratively
 *  \f[
 *  x_{n+1} = x_n - f(x_n)\frac{x_{n}-x_{n-1}}{f(x_{n}-f(x_{n-1})}
 *  \f]
 * in order to find the zero of the function f.
 *
 * Some details of the implementation :
 * - If at some step the root is bracketed, the method switch immediately
 * to the Brent's method.
 * - If at some step the method diverge, a small line search is performed.
 * - if no descent is found, the method return a NA number.
 *
 * @param f the function
 * @param x0 the first point of the algorithm
 * @param x1 the second point of the algorithm
 * @param tol the tolerance of the method
 * @return the zero of the function if the method converge, a NA value otherwise
**/
template <class Function>
Real SecantMethod( IFunction<Function> const& f, Real const& x0, Real const& x1, Real tol)
{
  Real a = x0, b = x1, fa = f(x0), fb = f(x1);
  // trivial cases
  if (fa==0.) return a;
  if (fb==0.) return b;
  // check if root is bracketed
  if (fa * fb <0.) { return BrentMethod(f, a, b, tol);}
  // set low and high values
  if (std::abs(fa) < std::abs(fb))
  {
    std::swap(a, b);
    std::swap(fa, fb);
  }
  // start secant iterations
  while( (std::abs(b-a)>tol))
  {
    Real delta = fb * (b - a) / (fb - fa);
    Real s = b - delta;
    // if we are out of bound, we stop
    if (s<f.xmin() || s > f.xmax()) return Arithmetic<Real>::NA();
    Real fs = f(s);
    // check if we can switch to Brent method
    if (fb * fs < 0) { return BrentMethod(f, b, s, tol); }
    // handle divergence
    if (std::abs(fs)>std::abs(fb))// divergence
    {
      bool dv = true;
      for (int i=0; i<16; i++)
      {
        delta /=2.;
        s = b - delta;
        fs = f(s);
        // check if we can switch to Brent method
        if (fb * fs < 0) { return BrentMethod(f, b, s, tol); }
        // ok !
        if (std::abs(fs)<std::abs(fb)) { break ; dv = false;}
      }
      // we were not lucky...
      if (dv) return Arithmetic<Real>::NA();
    }
    // update
    a =b; fa = fb;
    b =s; fb = fs;
    // trivial case
    if (fb==0.) return b;
  }
  return b;
}

/** @ingroup Analysis
 *  @brief perform the brent's algorithm.
 *
 * @param f the function
 * @param x0 the first point of the algorithm
 * @param x1 the second point of the algorithm
 * @param tol the tolerance of the method
 * @return the new value
**/
template <class Function>
Real BrentMethod( IFunction<Function> const& f, Real const& x0, Real const& x1, Real tol)
{
  Real ak = x0, bk = x1, fak = f(x0), fbk = f(x1);
  // trivial cases
  if (fak==0.) return ak;
  if (fbk==0.) return bk;
  // check if root is bracketed
  if (fak * fbk >0.) // secant method have to be used
  { return SecantMethod(f, ak, bk, tol);}
  // set low and high values
  if (std::abs(fak) < std::abs(fbk))
  {
    std::swap(ak, bk);
    std::swap(fak, fbk);
  }
  Real bkm1 = ak, fbkm1 = fak;
  bool mflag = true;
  Real bkm2 = bkm1; // value of bkm2 is not used at the first iteration as mflag = true

  int iter =0;
#ifdef STK_ANALYSIS_VERBOSE
  stk_cout << _T("iter = ") << iter <<_T(". bkm1, ak, bk = ") << bkm1 << _T(" ") << ak << _T(" ") << bk << _T("\n");
  stk_cout << _T("iter = ") << iter <<_T(". fbkm1, fak, fbk = ") << fbkm1 << _T(" ") << fak << _T(" ") << fbk << _T("\n");
#endif
  while( (std::abs(bk-ak)>tol) )
  {
    iter++;
    Real s;
    if ((fak != fbkm1) && (fbk != fbkm1)) // inverse quadratic approximation
    {
      Real dab = fak - fbk, dac = fak -fbkm1, dbc = fbk -fbkm1;
      s = ak * fbk * fbkm1 / dab / dac
        + bk * fak * fbkm1 / -dab / dbc
        + bkm1 * fak * fbk / dac / dbc;
      //stk_cout << _T("iter = ") << iter<<_T(". Quadratic. s=")<< s << _T("\n");
    }
    else // secante method
    {
      s = bk - fbk * (bk - ak) / (fbk - fak);
      //stk_cout << _T("iter = ") << iter<<_T(". Secant. s=")<< s << _T("\n");
    }
    // check if we shall fall-back to dichotomy
    Real tmp = (3. * ak + bk) / 4., diff1 =std::abs(bk - bkm1), diff2 = std::abs(bkm1-bkm2);
    if (!( ((s > tmp) && (s < bk)) || ((s < tmp) && (s > bk)) ) // s not between bk and tmp
       ||( mflag &&(  (std::abs(s-bk) >= diff1 / 2.)||(diff1 < tol)))
       ||(!mflag &&( (std::abs(s-bk) >= diff2/2.)||(diff2 < tol)))
       )
    {
       s = (ak + bk) / 2.;
       //stk_cout << _T("iter = ") << iter<<_T(". Dichotomy. s=")<< s << _T("\n");
       mflag = true;
    }
    else { mflag = false;}

    Real fs = f(s);
    bkm2 = bkm1; bkm1 = bk; fbkm1 = fbk;
    // check for bracketing the root
    if (fak * fs < 0) { bk = s; fbk = fs; }
    else              { ak = s; fak = fs; }
    // if |f(ak)| < |f(bk)| then swap (ak, bk)
    if (std::abs(fak) < std::abs(fbk))
    {
      std::swap(ak, bk);
      std::swap(fak, fbk);
    }
#ifdef STK_ANALYSIS_VERBOSE
    stk_cout << _T("iter = ") << iter <<_T(". bkm1, ak, bk = ") << bkm1 << _T(" ") << ak << _T(" ") << bk << _T("\n");
    stk_cout << _T("iter = ") << iter <<_T(". fbkm1, fak, fbk = ") << fbkm1 << _T(" ") << fak << _T(" ") << fbk << _T("\n");
#endif
    if (fbk==0.) return bk;
    if (iter > MAX_ITER) return Arithmetic<Real>::NA();
  }
  return bk;
}

/** @ingroup Analysis
 *  @brief find the zero of a function.
 *  Check if the initial values are inside the domain of definition of the
 *  function and call the Brent's method.
 *
 * @param f the function
 * @param x0 the first starting point of the algorithm
 * @param x1 the second starting point of the algorithm
 * @param tol the tolerance to apply
 * @return the zero of the function if any, a NA value otherwise
**/
template <class Function>
Real findZero( IFunction<Function> const& f, Real const& x0, Real const& x1, Real tol = Arithmetic<Real>:: epsilon())
{
  if (x0<f.xmin() || x0>f.xmax() || x1<f.xmin() || x1>f.xmax())
    STKDOMAIN_ERROR_2ARG(BrentMethod,x0,x1,An initial point is out of the domain);
  return BrentMethod(f, x0, x1, tol);
}

} // namesapce Law

} // namespace STK

#undef MAX_ITER
#endif /*STK_ALGO_FINDZERO_H*/

==> HDPenReg_0.89.9.tar.gz: HDPenReg/src/stkpp/projects/Analysis/include/STK_Algo_FindZero.h <==
/*--------------------------------------------------------------------*/
/*     Copyright (C) 2004-2007  Serge Iovleff

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place,
    Suite 330,
    Boston, MA 02111-1307
    USA

    Contact : S..._Dot_I..._At_stkpp_Dot_org (see copyright for ...)
*/

/*
 * Project:  stkpp::STatistiK
 * Purpose:  Method implementing Algorithms.
 * Author:   Serge Iovleff, S..._Dot_I..._At_stkpp_Dot_org (see copyright for ...)
 **/

/** @file STK_Algo.h
 *  @brief In this file we give some templated generic algorithms.
 **/

#ifndef STK_ALGO_FINDZERO_H
#define STK_ALGO_FINDZERO_H

#include "Sdk/include/STK_Macros.h"

#include "STKernel/include/STK_Integer.h"
#include "STKernel/include/STK_Real.h"

#include "STK_IFunction.h"

#define MAX_ITER 1000

namespace STK
{

namespace Algo
{
// forward declaration
template <class Function>
Real BrentMethod( IFunction<Function> const& f, Real const& x0, Real const& x1, Real tol);

/** @ingroup Analysis
 *  @brief apply the secant method for finding the zero of a function.
 *
 * Compute iteratively
 *  \f[
 *  x_{n+1} = x_n - f(x_n)\frac{x_{n}-x_{n-1}}{f(x_{n}-f(x_{n-1})}
 *  \f]
 * in order to find the zero of the function f.
 *
 * Some details of the implementation :
 * - If at some step the root is bracketed, the method switch immediately
 * to the Brent's method.
 * - If at some step the method diverge, a small line search is performed.
 * - if no descent is found, the method return a NA number.
 *
 * @param f the function
 * @param x0 the first point of the algorithm
 * @param x1 the second point of the algorithm
 * @param tol the tolerance of the method
 * @return the zero of the function if the method converge, a NA value otherwise
**/
template <class Function>
Real SecantMethod( IFunction<Function> const& f, Real const& x0, Real const& x1, Real tol)
{
  Real a = x0, b = x1, fa = f(x0), fb = f(x1);
  // trivial cases
  if (fa==0.) return a;
  if (fb==0.) return b;
  // check if root is bracketed
  if (fa * fb <0.) { return BrentMethod(f, a, b, tol);}
  // set low and high values
  if (std::abs(fa) < std::abs(fb))
  {
    std::swap(a, b);
    std::swap(fa, fb);
  }
  // start secant iterations
  while( (std::abs(b-a)>tol))
  {
    Real delta = fb * (b - a) / (fb - fa);
    Real s = b - delta;
    // if we are out of bound, we stop
    if (s<f.xmin() || s > f.xmax()) return Arithmetic<Real>::NA();
    Real fs = f(s);
    // check if we can switch to Brent method
    if (fb * fs < 0) { return BrentMethod(f, b, s, tol); }
    // handle divergence
    if (std::abs(fs)>std::abs(fb))// divergence
    {
      bool dv = true;
      for (int i=0; i<16; i++)
      {
        delta /=2.;
        s = b - delta;
        fs = f(s);
        // check if we can switch to Brent method
        if (fb * fs < 0) { return BrentMethod(f, b, s, tol); }
        // ok !
        if (std::abs(fs)<std::abs(fb)) { break ; dv = false;}
      }
      // we were not lucky...
      if (dv) return Arithmetic<Real>::NA();
    }
    // update
    a =b; fa = fb;
    b =s; fb = fs;
    // trivial case
    if (fb==0.) return b;
  }
  return b;
}

/** @ingroup Analysis
 *  @brief perform the brent's algorithm.
 *
 * @param f the function
 * @param x0 the first point of the algorithm
 * @param x1 the second point of the algorithm
 * @param tol the tolerance of the method
 * @return the new value
**/
template <class Function>
Real BrentMethod( IFunction<Function> const& f, Real const& x0, Real const& x1, Real tol)
{
  Real ak = x0, bk = x1, fak = f(x0), fbk = f(x1);
  // trivial cases
  if (fak==0.) return ak;
  if (fbk==0.) return bk;
  // check if root is bracketed
  if (fak * fbk >0.) // secant method have to be used
  { return SecantMethod(f, ak, bk, tol);}
  // set low and high values
  if (std::abs(fak) < std::abs(fbk))
  {
    std::swap(ak, bk);
    std::swap(fak, fbk);
  }
  Real bkm1 = ak, fbkm1 = fak;
  bool mflag = true;
  Real bkm2 = bkm1; // value of bkm2 is not used at the first iteration as mflag = true

  int iter =0;
#ifdef STK_ANALYSIS_VERBOSE
  stk_cout << _T("iter = ") << iter <<_T(". bkm1, ak, bk = ") << bkm1 << _T(" ") << ak << _T(" ") << bk << _T("\n");
  stk_cout << _T("iter = ") << iter <<_T(". fbkm1, fak, fbk = ") << fbkm1 << _T(" ") << fak << _T(" ") << fbk << _T("\n");
#endif
  while( (std::abs(bk-ak)>tol) )
  {
    iter++;
    Real s;
    if ((fak != fbkm1) && (fbk != fbkm1)) // inverse quadratic approximation
    {
      Real dab = fak - fbk, dac = fak -fbkm1, dbc = fbk -fbkm1;
      s = ak * fbk * fbkm1 / dab / dac
        + bk * fak * fbkm1 / -dab / dbc
        + bkm1 * fak * fbk / dac / dbc;
      //stk_cout << _T("iter = ") << iter<<_T(". Quadratic. s=")<< s << _T("\n");
    }
    else // secante method
    {
      s = bk - fbk * (bk - ak) / (fbk - fak);
      //stk_cout << _T("iter = ") << iter<<_T(". Secant. s=")<< s << _T("\n");
    }
    // check if we shall fall-back to dichotomy
    Real tmp = (3. * ak + bk) / 4., diff1 =std::abs(bk - bkm1), diff2 = std::abs(bkm1-bkm2);
    if (!( ((s > tmp) && (s < bk)) || ((s < tmp) && (s > bk)) ) // s not between bk and tmp
       ||( mflag &&(  (std::abs(s-bk) >= diff1 / 2.)||(diff1 < tol)))
       ||(!mflag &&( (std::abs(s-bk) >= diff2/2.)||(diff2 < tol)))
       )
    {
       s = (ak + bk) / 2.;
       //stk_cout << _T("iter = ") << iter<<_T(". Dichotomy. s=")<< s << _T("\n");
       mflag = true;
    }
    else { mflag = false;}

    Real fs = f(s);
    bkm2 = bkm1; bkm1 = bk; fbkm1 = fbk;
    // check for bracketing the root
    if (fak * fs < 0) { bk = s; fbk = fs; }
    else              { ak = s; fak = fs; }
    // if |f(ak)| < |f(bk)| then swap (ak, bk)
    if (std::abs(fak) < std::abs(fbk))
    {
      std::swap(ak, bk);
      std::swap(fak, fbk);
    }
#ifdef STK_ANALYSIS_VERBOSE
    stk_cout << _T("iter = ") << iter <<_T(". bkm1, ak, bk = ") << bkm1 << _T(" ") << ak << _T(" ") << bk << _T("\n");
    stk_cout << _T("iter = ") << iter <<_T(". fbkm1, fak, fbk = ") << fbkm1 << _T(" ") << fak << _T(" ") << fbk << _T("\n");
#endif
    if (fbk==0.) return bk;
    if (iter > MAX_ITER) return Arithmetic<Real>::NA();
  }
  return bk;
}

/** @ingroup Analysis
 *  @brief find the zero of a function.
 *  Check if the initial values are inside the domain of definition of the
 *  function and call the Brent's method.
 *
 * @param f the function
 * @param x0 the first starting point of the algorithm
 * @param x1 the second starting point of the algorithm
 * @param tol the tolerance to apply
 * @return the zero of the function if any, a NA value otherwise
**/
template <class Function>
Real findZero( IFunction<Function> const& f, Real const& x0, Real const& x1, Real tol = Arithmetic<Real>:: epsilon())
{
  if (x0<f.xmin() || x0>f.xmax() || x1<f.xmin() || x1>f.xmax())
    STKDOMAIN_ERROR_2ARG(BrentMethod,x0,x1,An initial point is out of the domain);
  return BrentMethod(f, x0, x1, tol);
}

} // namesapce Law

} // namespace STK

#undef MAX_ITER
#endif /*STK_ALGO_FINDZERO_H*/
